-- License CC0 (Creative Commons license) (c) darkfrei, 2022
-- push-blocks
local path = 'assets/players/'
local pb = {}
local function setBlockOutline (block)
	local d = 1/20
	local m = {} -- map of outlines
	local tiles = block.tiles -- list of tiles as {x1,y1, x2,y2, x3,y3 ...}
	for i = 1, #tiles, 2 do
		local x, y = tiles[i], tiles[i+1]
		if not m[y] then 
			m[y] = {} 
		end
		if not m[y][x] then 
			m[y][x] = true
		end
	end
	local lines = {}
	for y, xs in pairs (m) do
		for x, tabl in pairs (xs) do
			local h1 = m[y-1] and m[y-1][x]
			local v1 = m[y]   and m[y][x-1]
			local h2 = m[y+1] and m[y+1][x]
			local v2 = m[y]   and m[y][x+1]
			
			if not h1 then
				table.insert (lines, {x+d,y+d, x+1-d, y+d})
			end
			if not h2 then
				table.insert (lines, {x+d,y+1-d, x+1-d, y+1-d})
			end			
			if not v1 then
				table.insert (lines, {x+d,y+d, x+d, y+1-d})
			end
			if not v2 then
				table.insert (lines, {x+1-d,y+d, x+1-d, y+1-d})
			end

		end
	end
	block.lines = lines
end

local function remapTiles (block)
	local x0, y0 = block.x, block.y
	local dx, dy = 0, 0
	local tiles = block.tiles
	for i = 1, #tiles, 2 do
		local x, y = tiles[i], tiles[i+1]
		if x < dx then dx = x end
		if y < dy then dy = y end
	end
	block.x, block.y = block.x - dx, block.y - dy
	local w, h = 0, 0
	for i = 1, #tiles, 2 do
		local x, y = tiles[i]+dx, tiles[i+1]+dy
		tiles[i] = x
		tiles[i+1] = y
		if x > w then w = x end
		if y > h then h = y end
	end
	block.w = w+1
	block.h = h+1
end

function pb:load (level)
	local width, height = love.graphics.getDimensions()
	self.map = level.map
	self.gridWidth = level.w
	self.gridHeight = level.h
	self.gridSize = math.floor(math.min(width/(level.w), height/(level.h)))
	self.blocks = level.blocks
	for i, block in ipairs (self.blocks) do
		remapTiles (block)
		setBlockOutline (block)
	end
	self.agents = level.agents
	for i, agent in ipairs (self.agents) do
		setBlockOutline (agent)
		local filename = path..agent.name..'.png'
		local info = love.filesystem.getInfo(filename)
		if info then
			local image = love.graphics.newImage(filename)
			agent.image = image
			local width, height = image:getDimensions( )
			local scale = self.gridSize/(width/agent.w)
			agent.image_scale = scale
			agent.image_sx = scale
			agent.image_sy = scale
			agent.image_ox = image:getWidth()/2
			agent.image_oy = image:getHeight()/2
			agent.image_dx = agent.image_ox*agent.image_scale
			agent.image_dy = agent.image_oy*agent.image_scale
		end
	end
	self.activeAgentIndex = 1
	self.agent = self.agents[self.activeAgentIndex]
	self.agent.active = true
end

function pb:switchAgent ()
	self.agent.active = false
	local index = self.activeAgentIndex + 1
	if index > #self.agents then
		index = 1
	end
	self.activeAgentIndex = index
	self.agent = self.agents[self.activeAgentIndex]
	self.agent.active = true
end
local function isValueInList (value, list)
	for i, element in ipairs (list) do
		if element == value then 
			return true 
		end
	end
	return false
end

function pb:isBlockToMapCollision (block, dx, dy)
	local x, y = block.x, block.y
	local map = self.map
	for i = 1, #block.tiles-1, 2 do
		local mapX = x + block.tiles[i]	+ dx
		local mapY = y + block.tiles[i+1] + dy
		if map[mapY] and map[mapY][mapX] then 
			return true 
		end
	end
end

function pb:isBlockToBlockCollisionRough (blockA, blockB, dx, dy)
	local x1, y1 = blockA.x + dx, blockA.y + dy
	local h1, w1 = blockA.h, blockA.w
	local x2, y2 = blockB.x, blockB.y
	local h2, w2 = blockB.h, blockB.w
	if x1 < x2+w2 and x2 < x1+w1 and
		y1 < y2+h2 and y2 < y1+h1 then
		return true	
	end
	return false
end

function pb:isBlockToBlockCollisionFine (blockA, blockB, dx, dy)
	-- fine tile to tile collision detection
	-- check if blockA moves to dx, dy an collides with blockB
	local xA, yA = blockA.x + dx, blockA.y + dy
	local xB, yB = blockB.x, blockB.y
	local tilesA = blockA.tiles
	local tilesB = blockB.tiles
	for i = 1, #tilesA-1, 2 do
		local dXA, dYA = tilesA[i], tilesA[i+1]
		for j = 1, #tilesB-1, 2 do
			local dXB, dYB = tilesB[j], tilesB[j+1]
			if (xA+dXA == xB+dXB) and (yA+dYA == yB+dYB) then
				-- same x AND same y means collision
				return true
			end
		end
	end
	return false
end

function pb:isBlockToBlockCollision (blockA, blockB, dx, dy)
	-- check if blockA moves to dx, dy an collides with blockB
	-- rough
	if not self:isBlockToBlockCollisionRough (blockA, blockB, dx, dy) then
		return false
	elseif self:isBlockToBlockCollisionFine (blockA, blockB, dx, dy) then
	-- fine
		return true
	end
	return false
end


--testing

function pb:getCollisionExit ()
	-- all agents
	for i, agent in ipairs (self.agents) do
		if self:isCollisionAgentToAnyExitArea (agent) then
			return true -- agent collision with exit
		end
	end
end

function pb:getCollisionBlocks (blockA, blocks, dx, dy)
	-- agent to map or block to map collision
	-- deep function
	-- recursive detects all blocks to move, pushed by blockA;
	-- returns nothing if the move is not possible

	if self:isBlockToMapCollision (blockA, dx, dy) then
		return false
	end
	for i, agent in ipairs (self.agents) do
		if agent == self.agent then
			-- no collision detection with active agent
		elseif self:isBlockToBlockCollision (blockA, agent, dx, dy) then
			return false -- cannot move any agent
		end
	end
	for i, block in ipairs (self.blocks) do
		if block == blockA then
			-- self collision: do nothing
		elseif isValueInList (block, blocks) then
			-- block is already in list: do nothing
		elseif self:isBlockToBlockCollision (blockA, block, dx, dy) then
			-- checks if the agent is strong
			if block.heavy and not self.agent.heavy then
				--script that trigger dialogs when small fish try to push the pipe in the first level
				return false
			end
			table.insert (blocks, block)
			-- make it deeper!
			if not self:getCollisionBlocks (block, blocks, dx, dy) then
				-- if one of deep blocks cannot be moved
				return false
			end
		end
	end
	return true
end

function pb:getBlocksToMove (agent, dx, dy)
	local blocks = {}
	local canMove = self:getCollisionBlocks (agent, blocks, dx, dy)
	return blocks, canMove
end

function pb:moveBlock (block, dx, dy)
	block.x = block.x + dx
	block.y = block.y + dy
end

function pb:moveBlocks (blocks, dx, dy)
	for i, block in ipairs (blocks) do
		self:moveBlock (block, dx, dy)
	end
end

function pb:moveMetablocks (metablocks, dx, dy)
	for i, blocks in ipairs (metablocks) do
		self:moveBlocks (blocks, dx, dy)
	end
end


function pb:isCollisionBlockToAllBlocks (blockA, dx, dy)
	for i, block in ipairs (self.blocks) do
		if (block == blockA) then
			-- do nothing
		elseif self:isBlockToBlockCollision (blockA, block, dx, dy) then
			return true
		end
	end
	return false
end

function pb:isCollisionBlockToAllBlocksExcept (blockA, dx, dy, exceptionBlocks)
	for i, block in ipairs (self.blocks) do
		if (block == blockA) then
			-- do nothing
		elseif isValueInList (block, exceptionBlocks) then
			-- do nothing
		elseif self:isBlockToBlockCollision (blockA, block, dx, dy) then
			return true
		end
	end
	return false
end


function pb:isCollisionBlockToAllAgents (blockA, dx, dy)
--	(if blockA moves by dx dy)
	for i, agent in ipairs (self.agents) do
		if self:isBlockToBlockCollision (blockA, agent, dx, dy) then
			return agent -- dead agent :(
		end
	end
	return false
end

function pb:isCollisionAgentToAnyExitArea (agent)
	-- exit area is a special block
	for i, area in ipairs (self.exitAreas) do
		if self:isBlockToBlockCollision (area, agent, 0, 0) then 
-- dx and dy are 0
			return true -- collision one of areas to agent in /this/ position
		end
	end
	return false -- this agent has no collision with any of exit areas
end

function pb:areAllAgentsInExitAreas ()
	for i, agent in ipairs (self.agents) do
		if not self:isCollisionAgentToAnyExitArea (agent) then
			return false -- at least one of agents is not on exit, the game is not over
		end
		return true -- all agents are in exit areas, level is done
	end
end


local function isBlockInMetablocks (block, metablocks)
--	metablock is list of blocks
	for i, metablock in ipairs (metablocks) do
		if isValueInList (block, metablock) then
			return true
		end
	end
end


function pb:getCollisionBlocksFlat (blockA, dx, dy)
	local blocks = {}
	for i, blockB in ipairs (self.blocks) do 
		if isValueInList (blockB, blocks) then
			-- do nothing
		elseif blockA == blockB then
			-- do nothing
		elseif self:isBlockToBlockCollision (blockA, blockB, dx, dy) then
			table.insert (blocks, blockB)
		end
	end
	return blocks
end


function pb:metablockCreation (blockA, metablock)
	-- falling
	local dx, dy = 0, 1
	if self:isBlockToMapCollision (blockA, dx, dy) then 
		-- not falling :)
		blockA.text = "m1"
		return false 
	end
	if self:isCollisionBlockToAllAgents (blockA, dx, dy) then 
		-- not falling, on the fish
		blockA.text = "f1"
		return false 
	end
	-- neighbour collision blocks
	local blocks = pb:getCollisionBlocksFlat (blockA, dx, dy)
		
	for i, blockB in ipairs (blocks) do
		if isValueInList (blockB, metablock) then
--			-- do nothing
		else
			table.insert (metablock, blockB)
			if not self:metablockCreation (blockB, metablock) then
--				blockA.text = "d1"
				return false
			end
		end
	end
	return metablock
end

function pb:newMetablock (block)
	local metablock = {block}
	-- returns falling metablocks only
	if self:metablockCreation (block, metablock) then
		return metablock
	end
end

function pb:fallBlocks ()
-- no horizontal speed, but positive (down) vertical speed
	local dx, dy = 0, 1
	local metablocks = {}
	for i, block in ipairs (self.blocks) do
		block.text = nil
	end
	for i, block in ipairs (self.blocks) do
		if not isBlockInMetablocks (block, metablocks) then
			-- new metablock
			local metablock = self:newMetablock (block)
			if metablock then
				table.insert(metablocks, metablock)
			end
		end
	end
--	print ('#metablocks', #metablocks)
	
	for i = 1, self.gridHeight do
		local somethingMoved = false
		for iMetablock = #metablocks, 1, -1 do -- backwards
			local metablock = metablocks[iMetablock]
--			print ('metablock', iMetablock, 'blocks:', #metablock)
			local metablockFalling = true
			
			for j, block in ipairs (metablock) do
				if self:isBlockToMapCollision (block, dx, dy) then
					-- not falling
					block.text = "map"
					metablockFalling = false
				elseif self:isCollisionBlockToAllBlocksExcept (block, dx, dy, metablock) then
					-- collision to block: not falling
					block.text = "fest"
					metablockFalling = false
				elseif self:isCollisionBlockToAllBlocks (block, dx, dy) then
					-- collision to metablock
					block.text = "fallM"
				elseif self:isCollisionBlockToAllAgents (block, dx, dy) then
					local deadAgent = self:isCollisionBlockToAllAgents (block, dx, dy)
					block.text = 'on dead'
					deadAgent.dead = true
					metablockFalling = false
				else
					block.text = "fallS"
--					block moved only if metablock moved
					
				end
			end
			if metablockFalling then
				self:moveBlocks (metablock, dx, dy)
				somethingMoved = true
			else
				-- the metablock is not falling anymore
				table.remove(metablocks, iMetablock)
			end
		end
		if not somethingMoved then 
			-- nothing to move
			return 
		end
	end
end


function pb:mainMoving (dx, dy,dt)
	local agent = self.agent -- active agent
	if dx > 0 then 
		agent.direction = "right"
	elseif dx < 0 then
		agent.direction = "left"
	end
	local blocks, canMove = self:getBlocksToMove (agent, dx, dy)
	if canMove then
		self:moveBlock (agent, dx, dy) -- agent is block too
		self:moveBlocks (blocks, dx, dy)
--		self:fallBlocks (self.blocks, dx, dy) 
		self:fallBlocks ( ) 
	end
end

function pb:keypressedMoving (scancode)
	if scancode == 'w' or scancode == 'a' or scancode == 's' or scancode == 'd' then
		-- d means 1; a means -1; otherwise 0
		local dx = scancode == 'd' and 1 or scancode == 'a' and -1 or 0
		-- s means 1; w means -1; otherwise 0
		local dy = scancode == 's' and 1 or scancode == 'w' and -1 or 0
		pb:mainMoving (dx, dy)
	end
	if scancode == 'right' or scancode == 'left' or scancode == 'up' or scancode == 'down' then
		local dx = scancode == 'right' and 1 or scancode == 'left' and -1 or 0
		local dy = scancode == 'down' and 1 or scancode == 'up' and -1 or 0
		pb:mainMoving (dx, dy)
	end
end



-- please move the love.gamepadpressed to main.lua; 
-- here must be pb:gamepadpressed
-- / darkfrei

function love.gamepadpressed(joystick, button)
	local dx, dy = 0, 0 -- must be local
	if joystick:isGamepadDown("dpleft", "dpright", "dpup", "dpdown") then
		-- moving
		local dpleft = joystick:isGamepadDown("dpleft")
		local dpright = joystick:isGamepadDown("dpright")
		local dpup = joystick:isGamepadDown("dpup")
		local dpdown = joystick:isGamepadDown("dpdown")
		if dpleft then
			dx, dy = -1, 0
		elseif dpright then
			dx, dy = 1, 0
		elseif dpup then
			dx, dy = 0, -1
		elseif dpdown then
			dx, dy = 0, 1
		end
		pb:mainMoving (dx, dy)
		return -- nothing to do here
	end
	-- not moving:
	if joystick:isGamepadDown("a")	then
		pb:switchAgent ()
	elseif joystick:isGamepadDown("b")	then
		Talkies.clearMessages() 
		talkies=false
	elseif joystick:isGamepadDown("y")	then
		-- if helpison is boolean then just
		-- "helpison = not helpison"
		if helpison=="yes" then 
			helpison="no"
		elseif helpison=="no" then 
			helpison="yes"
		end
	elseif joystick:isGamepadDown("start")	then
		gamestatus="levelselection"
	elseif joystick:isGamepadDown("back")	then
		shader2=false
		gamestatus="options"
	end
		
	if joystick:isGamepad() then
		isgamepad=true
	end
end


---------------------------------------------------------------------------------------------------
-- draw
---------------------------------------------------------------------------------------------------

function pb:drawBackgroundGrid ()
	local gridSize = self.gridSize
	local gridWidth = self.gridWidth
	local gridHeight = self.gridHeight
	love.graphics.setLineWidth(1)
	love.graphics.setColor(0.3,0.4,0.4)
	for i = 0, gridWidth do
		love.graphics.line (i*gridSize, 0, i*gridSize, gridHeight*gridSize)
	end
	for i = 0, gridHeight do
		love.graphics.line (0, i*gridSize, gridWidth*gridSize, i*gridSize)
	end
end

function pb:drawMap ()
	local map = self.map
	local tileSize = self.gridSize
	love.graphics.setLineWidth(2)
	for y, xs in ipairs (map) do
		for x, value in ipairs (xs) do
			-- value 
			-- is boolean: true or false
			-- or string
			if value==true then 
				-- map tile
				-- beware of -1
				love.graphics.setColor(cs2tile1)
				love.graphics.rectangle ('fill', (x-1)*tileSize, (y-1)*tileSize, tileSize, tileSize)
				love.graphics.setColor(cs2tile2)
				love.graphics.rectangle ('fill', (x-1)*tileSize, (y-1)*tileSize, tileSize, tileSize-15)
			elseif value=="exit" then 
				-- map tile
				love.graphics.rectangle ('line', (x-1)*tileSize, (y-1)*tileSize, tileSize, tileSize-15)
				--borders
			elseif value=="border" then 
				-- map tile
				love.graphics.setColor(cs2border1)
				love.graphics.rectangle ('line', (x-1)*tileSize, (y-1)*tileSize, tileSize, tileSize)	
				love.graphics.rectangle ('fill', (x-1)*tileSize, (y-1)*tileSize, tileSize, tileSize)
				love.graphics.setColor(cs2border2)
				love.graphics.rectangle ('line', (x-1)*tileSize, (y-1)*tileSize, tileSize, tileSize-15)
				love.graphics.rectangle ('fill', (x-1)*tileSize, (y-1)*tileSize, tileSize, tileSize-15)
			end
		end
	end
	if extradebug==true and debugmode=="yes" then
		love.graphics.setFont(love.graphics.newFont(20))
		love.graphics.setColor(1,1,1)
		love.graphics.print("Tilesize :" .. tileSize ,50,100)
		love.graphics.print("Gridsize :" .. self.gridWidth .. "x " .. self.gridHeight,50,150)
	end
end


function pb:drawOutline	(block)
	local lines = block.lines
	local tileSize = self.gridSize
	local x, y = block.x-1, block.y-1
	--local exceptions = {1, 2, 3, 4, 6, 20, 21, 26, 28, 29, 30, 31, 48, 50, 52, 53, 54, 55, 56, 59, 60, 61}
	local exceptions = {}
	local exception = isValueInList (nLevel, exceptions) -- bool
	--do not draw Outlines if levels are:
	if objectscells=="yes" and not exception then
		for i, line in ipairs (lines) do
			love.graphics.line ((x+line[1])*tileSize, (y+line[2])*tileSize, (x+line[3])*tileSize, (y+line[4])*tileSize)
		end
	end
end


function pb:drawBlock (block)
--	if objectscells=="yes" then
--		love.graphics.setColor(cs2block)
--	end
	love.graphics.setColor(0.5,0.5,0.5)
	local x, y = block.x, block.y
	local tileSize = self.gridSize
	for i = 1, #block.tiles-1, 2 do
		local dx, dy = block.tiles[i], block.tiles[i+1]
		-- beware of -1
		love.graphics.rectangle ('fill', (x+dx-1)*tileSize, (y+dy-1)*tileSize, tileSize, tileSize) 
	end
	if block.text then
		love.graphics.setColor(1,1,1)
		love.graphics.print (block.text, (x-1)*tileSize, (y+1)*tileSize)
	end
end


function pb:drawBlocks ()
	love.graphics.setLineWidth(2)
	love.graphics.setColor(1,1,0.5)
	for i, block in ipairs (self.blocks) do
		-- draw filled block
		love.graphics.setLineWidth(1)
		--love.graphics.setColor(1,1,0.5)
		self:drawBlock (block)
		-- outline
		love.graphics.setLineWidth(3)
		if block.heavy then
			love.graphics.setColor(0,1,1)
		else
			love.graphics.setColor(0,1,0)
		end
		self:drawOutline (block)
	end
end


function pb:drawDeadAgent (agent)
	local tileSize = self.gridSize 
	local x = (agent.x-1)*tileSize
	local y = (agent.y-1)*tileSize
	local w = agent.w*tileSize
	local h = agent.h*tileSize
	love.graphics.line (x, y, x+w, y+h)
	love.graphics.line (x, y+h, x+w, y)
end

local function drawTexture (agent, tileSize)
	local name = agent.name
	local x = (agent.x-1)*tileSize+agent.image_dx
	local y = (agent.y-1)*tileSize+agent.image_dy
	local sx = agent.image_sx
	if agent.direction and agent.direction == "left" then
		sx = -agent.image_sx
	end
	local sy = agent.image_sy
	local ox = agent.image_ox
	local oy = agent.image_oy
	local images = {
		['fish-3x1']=c64fish3x1,
		['fish-4x2']=c64fish4x2,
		}
	local image = images[name]

	--c64 (pallete 2)
	love.graphics.draw (image, x, y, 0, sx, sy, ox, oy)
end

local  function drawagentrectangle(agent, tileSize)
	local sx = agent.image_sx
	love.graphics.rectangle ('fill', (agent.x-3)*tileSize+agent.image_dx, (agent.y-2)*tileSize+agent.image_dy, agent.image_dx*2, agent.image_dy*2)
end


function pb:drawAgents ()
	local activeAgent = self.agent
	local tileSize = self.gridSize
	-- make it local
	local gridSize=self.gridSize
	for i, agent in ipairs (self.agents) do
		if agent.image then
			-- istead of
--			cs2agent1={0.313,0.270,0.607}
--			cs2agent2={0.623,0.305,0.266}
			-- nice colors are
			local cs2agent1={.9,.9,.8}
			local cs2agent2={0.4,0.4,0.3}

			if agent == activeAgent then
				love.graphics.setColor(cs2agent1)
				drawTexture (agent, tileSize)
			else
				love.graphics.setColor(cs2agent2)
				drawTexture (agent, tileSize)
			end
		else
			if agent == activeAgent then
				love.graphics.setColor(1,1,1)
				self:drawBlock (agent)
				local x, y = agent.x, agent.y
				love.graphics.setColor (0, 0, 0)
				love.graphics.print (agent.x..' '..agent.y, (agent.x-1)*tileSize, (agent.y-1)*tileSize)
			else
				love.graphics.setColor(0.75,0.75,0.5)
				self:drawBlock (agent)
			end
			-- outline
			love.graphics.setLineWidth(3)
			if agent.heavy then
				love.graphics.setColor(0,1,1)
			else
				love.graphics.setColor(0,1,0)
			end
			self:drawOutline	(agent)
		end
		if agent.dead then
			love.graphics.setColor(0,0,0)
			self:drawDeadAgent (agent)
		end
	end
	love.graphics.setColor(1,1,1)
end


function pb:drawMouse ()
	local mx, my = love.mouse.getPosition()
	local tileSize = self.gridSize 
	local x = math.floor(mx/tileSize)+1
	local y = math.floor(my/tileSize)+1
end
return pb
